/**
 * A Low Energy radio stack for the NRF microcontrollers.
 *
 * @author  Ehsan Tofigh
 * @project lcx
 * @file    src/libradio/ale.c
 * @license
 */

#include <libradio/ale.h>
#include <libcortex/nvic.h>
#include <libcortex/radio.h>
#include <libcortex/rtc.h>
#include <libcortex/rng.h>
#include <libcortex/ppi.h>
#include <libcortex/gpiote.h>
#include <libcortex/util.h>

#include <string.h>

/** Maximum payload size in bytes. */
#define MAX_PAYLOAD_SIZE   251U

/** Access address used for the channel. */
#define ACCESS_ADDRESS     0x41424347U

/**
 * The pre-programmed PPI channel for triggering AES Key generation when the
 * radio is ready.
 */
#define PPI_RADIO_READY_KSGEN   24U

/**
 * The pre-programmed PPI channel for triggering AES decryption when the address
 * is received on the radio.
 */
#define PPI_RADIO_ADDRESS_CRYPT 25U

/** ALE packet structures. */
struct packet {
	struct lrd_ale_header header;          /**< Packet header. */
	union {
		uint8_t raw[MAX_PAYLOAD_SIZE]; /**< Raw view of the packet. */
	};
} __attribute__((packed));

/** ALE state structure. */
static struct {
	struct packet tmp_packet;            /**< Buffer to temporarily keep the packet for encryption/decryption. */
	struct packet rx_packet;             /**< Buffer to keep the decrypted packet. */
} ale;

/** Packet structure configuration. */
static const struct lcx_radio_packet_config packet_config = {
	.pcnf0 = {
		.PLEN    = lcx_radio_pcnf0_plen_map[LCX_RADIO_PCNF0_PLEN_LONGRANGE],
		.LFLEN   = 8U,
		.S0LEN   = 1U,
		.S1LEN   = 8U,
		.S1INCL  = lcx_radio_pcnf0_s1incl_map[LCX_RADIO_PCNF0_S1INCL_INCLUDE],
		.CILEN   = 2U,
		.TERMLEN = 3U,
	},
	.pcnf1 = {
		.WHITEEN = lcx_radio_pcnf1_whiteen_map[LCX_RADIO_PCNF1_WHITEEN_ENABLED],
		.BALEN   = 3U, /* Base address length in number of bytes minus 1. */
		.MAXLEN  = MAX_PAYLOAD_SIZE + 4U,
	},
	.crccnf = {
		.LEN      = lcx_radio_crccnf_len_map[LCX_RADIO_CRCCNF_LEN_THREE],
		.SKIPADDR = lcx_radio_crccnf_skipaddr_map[LCX_RADIO_CRCCNF_SKIPADDR_SKIP],
	},
	.crcpoly = { .CRCPOLY = 0x000065BU },
	.crcinit = { .CRCINIT = 0x0555555U },
	.tifs    = { .TIFS    = 150U       }, /* Interframe spacing in us. */
};

/** List of ALE channel frequencies. */
static const uint16_t ch_freq[] = {
	[36U] = 2400U,
	[37U] = 2402U,
	[38U] = 2426U,
	[39U] = 2480U,
};

/**
 * Set the access address.
 *
 * @param[in] addr 4 byte access address.
 */
static void set_access_address (const uint32_t addr)
{
	LCX_RADIO->BASE0.BASE0       = (addr << 8U);
	LCX_RADIO->PREFIX0.AP0       = (uint8_t) (addr >> 24U);
	LCX_RADIO->RXADDRESSES.ADDR0 = 1U;
}

/**
 * Main radio interrupt handler.
 */
static void radio_irq_handler (void)
{
	if (lcx_radio_test_and_clear_event(LCX_RADIO_EVENTS_PHYEND)) {
	}

	if (lcx_radio_test_and_clear_event(LCX_RADIO_EVENTS_ADDRESS)) {
	}

	if (lcx_radio_test_and_clear_event(LCX_RADIO_EVENTS_CRCOK)) {
	}

	if (lcx_radio_test_and_clear_event(LCX_RADIO_EVENTS_CCABUSY)) {
	}
}

bool lrd_ale_init (const struct lrd_ale_handle *handle)
{
	lcx_assert(handle == NULL);

	/* Enable the radio: */
	lcx_assert(lcx_radio_init() == false);

	/* Fix for errata 191: */
	*(volatile uint32_t *) 0x40001740U =
		((*((volatile uint32_t *) 0x40001740U)) & 0x7FFF00FFU) | 0x80000000U | (((uint32_t)(196U)) << 8U);

	/* Configure the radio: */
	lcx_assert(lcx_radio_set_packet_config(&packet_config) == false);
	lcx_assert(lcx_radio_set_mode(LCX_RADIO_MODE_MODE_BLE_LR125KBIT) == false);
	lcx_assert(lcx_radio_set_freq(ch_freq[36U]) == false);
	LCX_RADIO->DATAWHITEIV.DATAWHITEIV = 36U;
	LCX_RADIO->TXADDRESS.TXADDRESS = 0U;

	/* Configure the cryptography engine: */
	lcx_assert(lcx_ccm_init(LCX_CCM_MODE_DATARATE_125KBPS, handle->key, handle->iv) == false);

	/* Set the access address: */
	set_access_address(ACCESS_ADDRESS);

	/* Configure the clear channel assessment: */
	LCX_RADIO->CCACTRL.CCAEDTHRES = 30U;

	/* Start the RNG: */
	lcx_rng_start();

	/* Enable the interrupts: */
	lcx_radio_enable_irq(LCX_RADIO_IRQ_PHYEND);
	lcx_radio_enable_irq(LCX_RADIO_IRQ_CCABUSY);
	lcx_nvic_enable_src(LCX_RADIO_NVIC_SRC, 1U, 0U, &radio_irq_handler);

	return true;
}

bool lrd_ale_start_receiver (void)
{
	/* Configure the encryption engine: */
	lcx_assert(lcx_ccm_start(LCX_CCM_MODE_MODE_DECRYPTION,
				 (const uint8_t *) &ale.tmp_packet,
				 (const uint8_t *) &ale.rx_packet) == false);
	lcx_assert(lcx_radio_set_ptr((const uint8_t *) &ale.tmp_packet) == false);
	lcx_ccm_trigger_task(LCX_CCM_TASKS_KSGEN);

	/* *************** This does not work! ***************** */
	lcx_assert(lcx_ppi_ch_enable(PPI_RADIO_ADDRESS_CRYPT, true) == false);

	/* ******************* This workS! ********************* */
	// lcx_ppi_ch_init(0U, lcx_radio_get_event(LCX_RADIO_EVENTS_PAYLOAD),
	//       lcx_ccm_get_task(LCX_CCM_TASKS_CRYPT), NULL);
	// lcx_assert(lcx_ppi_ch_enable(0U, true) == false);

	/* Configure the shortcuts: */
	lcx_set_shorts(LCX_RADIO, .RXREADY_START = 1U, .PHYEND_DISABLE = 1U, .DISABLED_RXEN = 0U, .ADDRESS_RSSISTART = 1U, .DISABLED_RSSISTOP = 1U);

	lcx_radio_trigger_task(LCX_RADIO_TASKS_RXEN);

	return true;
}

uint8_t lrd_ale_read_edlevel (const uint32_t cnt)
{
	/* Set the number of measurements: */
	LCX_RADIO->EDCNT.EDCNT = cnt;

	/* Configure the shortcuts: */
	lcx_set_shorts(LCX_RADIO, .READY_EDSTART = 1U, .EDEND_DISABLE = 1U);

	lcx_radio_trigger_task(LCX_RADIO_TASKS_RXEN);
	while (lcx_radio_test_and_clear_event(LCX_RADIO_EVENTS_EDEND) == false);

	return (uint8_t) LCX_RADIO->EDSAMPLE.EDLVL;
}

bool lrd_ale_transmit (struct lrd_ale_header *header,
		       const size_t           len,
		       const bool             cca)
{
	lcx_assert(header == NULL);
	lcx_assert(LCX_RADIO->STATE.STATE != lcx_radio_state_state_map[LCX_RADIO_STATE_STATE_DISABLED]);

	/* Check the size of the payload: */
	size_t l = len - sizeof(*header);
	lcx_assert(l > MAX_PAYLOAD_SIZE);
	header->length = (uint8_t) l;

	/* Configure the encryption engine: */
	lcx_assert(lcx_ppi_ch_enable(PPI_RADIO_ADDRESS_CRYPT, false) == false);
	lcx_assert(lcx_ccm_start(LCX_CCM_MODE_MODE_ENCRYPTION,
				 (const uint8_t *) header,
				 (const uint8_t *) &ale.tmp_packet) == false);
	lcx_assert(lcx_radio_set_ptr((const uint8_t *) &ale.tmp_packet) == false);
	lcx_ccm_trigger_task(LCX_CCM_TASKS_KSGEN);

	/* Configure the shortcuts: */
	if (cca) {
		lcx_set_shorts(LCX_RADIO, .RXREADY_CCASTART = 1U, .TXREADY_START = 1U, .PHYEND_DISABLE = 1U, .CCAIDLE_TXEN = 1U, .CCABUSY_DISABLE = 1U);
		lcx_radio_trigger_task(LCX_RADIO_TASKS_RXEN);
	} else {
		lcx_set_shorts(LCX_RADIO, .TXREADY_START = 1U, .PHYEND_DISABLE = 1U);
		lcx_radio_trigger_task(LCX_RADIO_TASKS_TXEN);
	}

	return true;
}
